/*
convolve.c
Author: Hayden Walles
Date: 9 September 2007
Routines to handle the setup and performance of image convolutions.
*/

/*
    Copyright (C) 2007 Hayden Walles

    This file is part of Seamstress.

    Seamstress is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Seamstress is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "seams.h"


MATRIX *convolvePicture(SEAM_COMMON *pic,const MATRIX *base, const MATRIX *filter, int *error){
  MATRIX *result;
  int ci,cj;
  int i,j;
  int sum;
  int oidx,iidx,fidx;
  int startrow,startcol;

  int vborder,hborder;
  int filtsize;

  int foffsets[matrixCols(filter)*matrixRows(filter)];

  /*Find out how big the borders are around the image.  They are symmetrical in each direction.*/
  hborder=(matrixCols(base)-pic->width)/2;
  vborder=(matrixRows(base)-pic->height)/2;

  /*Precompute relative offsets for the input corresponding to each filter entry.*/
  filtsize=0;
  for(i=0;i<matrixRows(filter);i++){
    for(j=0;j<matrixCols(filter);j++){
      foffsets[filtsize]=i*matrixCols(base)+j;
      filtsize++;
    }
  }


  /*Grab the new matrix.  For the moment I will just always return an integer matrix here.*/
  result=newMatrix(pic->height,pic->width,MT_INTEGER,error);
  if(result==NULL){
    SET_ERROR(SEAMSTRESS_E_MEMORY);
    return NULL;
  }
  
  /*Check only integers are involved.*/

  if(filter->type!=MT_INTEGER){
    SET_ERROR(SEAMSTRESS_E_WRONGTYPE);
    return NULL;
  }
  
  /*That's the admin over with, now actually convolve them.  This is relatively easy, except we need to be careful around the edges.
    First, calculate a 'centre' element for the filter matrix.  This might not be the actual centre if an even dimension is involved.*/

  ci=filter->rows/2;
  cj=filter->cols/2;


  /*All we have to do is shift the filter over the base (though perhaps not the entire base). We start at the top left and head right, then start at the next row.  The first computation will give us the top left entry of the output matrix.  The border around the base is supposed to perfectly fit the actual image content plus any of the filters used.*/
  
  startrow = vborder-ci;

  startcol=hborder-cj;

  
  /*ix = x on input, ox = x on output, fx= x on filter.*/

  oidx=0;
  iidx=startrow*matrixCols(base)+startcol;

  for(i=0;i<pic->height;i++){
    for(j=0;j<pic->width;j++){
      sum=0;
      for(fidx=0;fidx<filtsize;fidx++){
	sum+=(base->buffer.integers[iidx+foffsets[fidx]])*filter->buffer.integers[fidx];
      }
      
      result->buffer.integers[oidx]=abs(sum);

      iidx++;
      oidx++;

    }
    iidx+=hborder*2;
  }
  
  
  
  SET_ERROR(SEAMSTRESS_E_OK);
  return result;


} 




MATRIX *dualConvolvePicture(SEAM_COMMON *pic,const MATRIX *base, const MATRIX *filter1,int norm1,const MATRIX *filter2,int norm2, int *error){
  MATRIX *result;
  int i,j;
  int sum1,sum2;
  int oidx,iidx1,iidx2,fidx;
  int startrow1,startcol1,startrow2,startcol2;

  int vborder,hborder;
  int filtsize1,filtsize2;

  int foffsets1[matrixCols(filter1)*matrixRows(filter1)],foffsets2[matrixCols(filter2)*matrixRows(filter2)];

  /*Find out how big the borders are around the image.  They are symmetrical in each direction.*/
  hborder=(matrixCols(base)-pic->width)/2;
  vborder=(matrixRows(base)-pic->height)/2;

  /*Precompute relative offsets for the input corresponding to each filter entry.*/
  filtsize1=0;
  for(i=0;i<matrixRows(filter1);i++){
    for(j=0;j<matrixCols(filter1);j++){
      foffsets1[filtsize1]=i*matrixCols(base)+j;
      filtsize1++;
    }
  }
  filtsize2=0;
  for(i=0;i<matrixRows(filter2);i++){
    for(j=0;j<matrixCols(filter2);j++){
      foffsets2[filtsize2]=i*matrixCols(base)+j;
      filtsize2++;
    }
  }

  /*Grab the new matrix.  For the moment I will just always return an integer matrix here.*/
  result=newMatrix(pic->height,pic->width,MT_INTEGER,error);
  if(result==NULL){
    SET_ERROR(SEAMSTRESS_E_MEMORY);
    return NULL;
  }
  
  /*Check only integers are involved.*/

  if((filter1->type!=MT_INTEGER)||(filter2->type!=MT_INTEGER)){
    SET_ERROR(SEAMSTRESS_E_WRONGTYPE);
    return NULL;
  }
  
  /*That's the admin over with, now actually convolve them.  This is relatively easy, except we need to be careful around the edges.
    First, calculate a 'centre' element for the filter matrix.  This might not be the actual centre if an even dimension is involved.*/


  /*All we have to do is shift the filter over the base (though perhaps not the entire base). We start at the top left and head right, then start at the next row.  The first computation will give us the top left entry of the output matrix.  The border around the base is supposed to perfectly fit the actual image content plus any of the filters used.*/
  
  startrow1 = vborder-filter1->rows/2;
  startrow2 = vborder-filter2->rows/2;

  startcol1=hborder-filter1->cols/2;
  startcol2=hborder-filter2->cols/2;

  
  /*ix = x on input, ox = x on output, fx= x on filter.*/
  oidx=0;
  iidx1=startrow1*matrixCols(base)+startcol1;
  iidx2=startrow2*matrixCols(base)+startcol2;
  for(i=0;i<pic->height;i++){
    for(j=0;j<pic->width;j++){
      sum1=0;
      for(fidx=0;fidx<filtsize1;fidx++){
	sum1+=(base->buffer.integers[iidx1+foffsets1[fidx]])*filter1->buffer.integers[fidx];
      }
      sum2=0;
      for(fidx=0;fidx<filtsize2;fidx++){
	sum2+=(base->buffer.integers[iidx2+foffsets2[fidx]])*filter2->buffer.integers[fidx];
      }
      if(norm1==norm2)
	result->buffer.integers[oidx]=(abs(sum1)+abs(sum2))/norm1/2;
      else
	result->buffer.integers[oidx]=(abs(sum1)/norm1+abs(sum2)/norm2)/2;
      iidx1++;
      iidx2++;
      oidx++;
    }
    iidx1+=hborder*2;
    iidx2+=hborder*2;
  }
  




  
  SET_ERROR(SEAMSTRESS_E_OK);
  return result;


} 


/*This is designed for incremental updates of the energy function.  It recomputes energy for pixels in each row between froms[row] and tos[row] inclusive.  It is specialised to compute a function of the form abs(f1*image)+abs(f2*image) at each point, where f1 and f2 are linear filters.  The reason this has to take an unpicker, rather than a common structure is that the nodes must be accessible. This is something that it would be nice to tidy up.*/
void dualConvolvePictureInc(SEAM_UNPICKER *pic,const MATRIX *base, const MATRIX *filter1,int norm1,const MATRIX *filter2,int norm2,int *froms, int *tos,MATRIX *result){
  int i,j;
  int sum1,sum2;
  int oidx,iidx1,iidx2,fidx;
  int startrow1,startcol1,startrow2,startcol2;

  int vborder,hborder;
  int filtsize1,filtsize2;
#ifdef HAVE_DYNAMIC_ARRAYS
  int foffsets1[matrixCols(filter1)*matrixRows(filter1)],foffsets2[matrixCols(filter2)*matrixRows(filter2)];
#elif defined(HAVE_ALLOCA)
  int *foffsets1,*foffsets2;
  foffsets1=(int *)alloca(sizeof(int)*matrixCols(filter1)*matrixRows(filter1));
  foffsets2=(int *)alloca(sizeof(int)*matrixCols(filter2)*matrixRows(filter2));
#else
  /*I should replace this with heap-allocated arrays, of course.*/
#error Need dynamic arrays or alloca!
#endif

  /*The most important thing to remember here is that we have to make all accesses to the energy map indirectly via the picture's nodes.  We are convolving _as though_ the picture had been shrunk by a certain number of seams already.  We can assume, though, that the luminance map has been shrunk correctly (that's to make sure the borders remain consistent)*/

  /*Find out how big the borders are around the image.  They are symmetrical in each direction.*/
  hborder=(matrixCols(base)-pic->common.width)/2;
  vborder=(matrixRows(base)-pic->common.height)/2;

  /*Precompute relative offsets for the input corresponding to each filter entry.*/
  filtsize1=0;
  for(i=0;i<matrixRows(filter1);i++){
    for(j=0;j<matrixCols(filter1);j++){
      foffsets1[filtsize1]=i*matrixCols(base)+j;
      filtsize1++;
    }
  }
  filtsize2=0;
  for(i=0;i<matrixRows(filter2);i++){
    for(j=0;j<matrixCols(filter2);j++){
      foffsets2[filtsize2]=i*matrixCols(base)+j;
      filtsize2++;
    }
  }

  /*Grab the new matrix.  For the moment I will just always return an integer matrix here.*/
  
  /*Check only integers are involved.*/

  if((filter1->type!=MT_INTEGER)||(filter2->type!=MT_INTEGER)){
    return;
  }
  
  /*That's the admin over with, now actually convolve them.  This is relatively easy, except we need to be careful around the edges.
    First, calculate a 'centre' element for the filter matrix.  This might not be the actual centre if an even dimension is involved.*/


  /*All we have to do is shift the filter over the base (though perhaps not the entire base). We start at the top left and head right, then start at the next row.  The first computation will give us the top left entry of the output matrix.  The border around the base is supposed to perfectly fit the actual image content plus any of the filters used.*/
  
  startrow1 = vborder-filter1->rows/2;
  startrow2 = vborder-filter2->rows/2;

  startcol1=hborder-filter1->cols/2;
  startcol2=hborder-filter2->cols/2;

  
  /*ix = x on input, ox = x on output, fx= x on filter.*/
  
  oidx=0;
  iidx1=startrow1*matrixCols(base)+startcol1;
  iidx2=startrow2*matrixCols(base)+startcol2;
  for(i=0;i<pic->common.height;i++){
    for(j=froms[i];j<=tos[i];j++){
      sum1=0;
      for(fidx=0;fidx<filtsize1;fidx++){
	sum1+=(base->buffer.integers[iidx1+j+foffsets1[fidx]])*filter1->buffer.integers[fidx];
      }
      sum2=0;
      for(fidx=0;fidx<filtsize2;fidx++){
	sum2+=(base->buffer.integers[iidx2+j+foffsets2[fidx]])*filter2->buffer.integers[fidx];
      }
      if(norm1==norm2)
	result->buffer.integers[oidx+pic->nodes[oidx+j].offset]=(abs(sum1)+abs(sum2))/norm1/2;
	else
      result->buffer.integers[oidx+pic->nodes[oidx+j].offset]=(abs(sum1)/norm1+abs(sum2)/norm2)/2;

    }
    iidx1+=matrixCols(base);
    iidx2+=matrixCols(base);
    oidx+=matrixCols(result);
  }
  




  
  


} 
