/*
filters.c
Author: Hayden Walles
Date: 9 September 2007
Functions to create filter matrices that are useful for computing energy functions.
*/

/*
    Copyright (C) 2007 Hayden Walles

    This file is part of Seamstress.

    Seamstress is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Seamstress is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "seams.h"

#define SQR(x) ((x)*(x))


/*This is just a helper that computes the transpose of the specified matrix.*/
MATRIX *transpose(MATRIX *in,int *error){
  MATRIX *out;
  int i,j;

  if(error)
    *error=SEAMSTRESS_E_OK;

  if((out=newMatrix(matrixCols(in),matrixRows(in),MT_INTEGER,error))==NULL){
    SET_ERROR(SEAMSTRESS_E_MEMORY);
    return NULL;
  }

  for(i=0;i<matrixRows(in);i++){
    for(j=0;j<matrixCols(in);j++)
      setInteger(out,j,i,getInteger(in,i,j));
  }

  
  return out;
}

MATRIX *createXSobel(int *error){
  MATRIX *new;

  if(error)
    *error=MATRIX_E_OK;
  
  new=newMatrix(3,3,MT_INTEGER,error);
  if(new==NULL)
    return NULL;

  setInteger(new,0,0,-1);
  setInteger(new,0,1,0);
  setInteger(new,0,2,1);
  setInteger(new,1,0,-2);
  setInteger(new,1,1,0);
  setInteger(new,1,2,2);
  setInteger(new,2,0,-1);
  setInteger(new,2,1,0);
  setInteger(new,2,2,1);

  return new;
}

MATRIX *createYSobel(int *error){
  MATRIX *new;

  if(error)
    *error=MATRIX_E_OK;
  
  new=newMatrix(3,3,MT_INTEGER,error);
  if(new==NULL)
    return NULL;

  setInteger(new,0,0,-1);
  setInteger(new,1,0,0);
  setInteger(new,2,0,1);
  setInteger(new,0,1,-2);
  setInteger(new,1,1,0);
  setInteger(new,2,1,2);
  setInteger(new,0,2,-1);
  setInteger(new,1,2,0);
  setInteger(new,2,2,1);

  return new;
}

/*This creates a filter which is the partial derivative wrt x of an isotropic Gaussian function with standard deviation sigma.*/
MATRIX *createXGaussianGradient(double sigma,int *norm,int *error){
  MATRIX *new;
  int i,j,width,height,shift,max,min;
  double rs,hoffset,voffset,x,y;
  int pnorm=0,nnorm=0;

  if(error)
    *error=MATRIX_E_OK;
  
  width=(int)(sigma*7.0);
  height=(int)(sigma*5.0);
  hoffset=(width-1.0)/2.0;
  voffset=(height-1.0)/2.0;

  /*The size of the matrix to use is estimated heuristically from
   the parameter sigma*/

  new=newMatrix(height,width,MT_INTEGER,NULL);
  if(new==NULL){
    SET_ERROR(SEAMSTRESS_E_MEMORY);
    return NULL;
  }
  
  max=-1;
  min=1;
  for(i=0;i<height;i++){
    for(j=0;j<width;j++){
      x=j-hoffset;
      y=i-voffset;
      rs=SQR(x)+SQR(y);
      setInteger(new,i,j,(int)(-209*x/sigma*exp(-rs/(2*SQR(sigma)))));
      if(getInteger(new,i,j)>0)
	pnorm+=getInteger(new,i,j);
      else
	nnorm+=abs(getInteger(new,i,j));
      if(getInteger(new,i,j)>max)
	max=getInteger(new,i,j);
      if(getInteger(new,i,j)<min)
	min=getInteger(new,i,j);
    }
      }

  *norm=MAX(pnorm,nnorm); /*actually, I know these should both be the same.*/

  return new;
  

}

